
from abc import ABC, abstractmethod
from typing import Dict
import numpy as np

class AbstractIterativeStatistics(ABC):
    """
        A basic abstract class to compute iteratives statistics.
    """
    def __init__(self, dim:int, state: object = None):
        self.dimension = dim
        if state is None :
            self.state = np.zeros(dim)
            self.iteration = 0
        else :
            self.load_from_state(state)
        

    @abstractmethod
    def increment(self, data : np.array):
        """
            An abstract method to implement. It must contain the algorithm to compute iteratively statistics.
        """
        raise Exception('Not implemented method')
    
    def save_state(self):
        """
            A method that save the current state.
        """
        return {'iteration': self.iteration, 'state': self.state}

    def load_from_state(self, state: object):
        """
            A method that load the current state
        """
        self.iteration = state.get('iteration', 0)
        self.state = state.get('state', None)

    def get_stats(self):
        """
            A method that return the current state
        """
        return self.state
    
    def get_iteration(self):
        """
            A method that return the current iteration value
        """
        return self.iteration


import numpy as np

from iterative_stats.sensitivity.abstract_sensitivity import IterativeAbstractSensitivity
from iterative_stats.iterative_dotproduct import IterativeDotProduct
from iterative_stats.utils.logger import logger
from iterative_stats.sensitivity import SALTELLI

class IterativeSensitivitySaltelli(IterativeAbstractSensitivity):
    """
    Estimates the Sobol indices iteratively based on the Saltelli formula.
    """
    def __init__(self, nb_parms: int, dim: int = 1, second_order: bool = False, state: object = None):
        super().__init__(nb_parms = nb_parms, dim = dim, 
                            second_order=second_order, name = SALTELLI, state=state)
        
        self.dotproduct_AE = [IterativeDotProduct(dim, iterative_shifted_mean_1 = self.iterative_shifted_mean_A, iterative_shifted_mean_2 = self.iterative_shifted_mean_E[i]) for i in range(self.nb_parms)]
        self.dotproduct_BE = [IterativeDotProduct(dim, iterative_shifted_mean_1 = self.iterative_shifted_mean_B, iterative_shifted_mean_2 = self.iterative_shifted_mean_E[i]) for i in range(self.nb_parms)]
        
       
    def _increment(self, data):
        sample_E = data[2:(2 + self.nb_parms)]
        for p in range(self.nb_parms):
            self.dotproduct_AE[p].increment(data[0],sample_E[p], shift = self.mean_tot.get_stats())
            self.dotproduct_BE[p].increment(data[1],sample_E[p], shift = self.mean_tot.get_stats())

        
    def _compute_varianceI(self):
        mean_A = self.iterative_shifted_mean_A.get_stats()
        mean_B = self.iterative_shifted_mean_B.get_stats()
        var = np.zeros((self.dimension, self.nb_parms))
        for p in range(self.nb_parms):
            var[:,p] = self.dotproduct_BE[p].get_stats() - mean_A * mean_B
        return var

    def _compute_VTi(self):
        vti = np.zeros((self.dimension, self.nb_parms))
        mean_A = self.iterative_shifted_mean_A.get_stats()
        for p in range(self.nb_parms):
            vti[:,p] = self.var_A.get_stats() - self.dotproduct_AE[p].get_stats() +  mean_A * mean_A
        return vti

    def getSecondOrderIndices(self) -> np.array:
        """
            Compute the self.nb_parms second order sensitivity indices
        """
        raise Exception('Not implemented method')

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()

        state['dotproduct_AE'] = [self.dotproduct_AE[i].save_state() for i in range(self.nb_parms)]
        state['dotproduct_BE'] = [self.dotproduct_BE[i].save_state() for i in range(self.nb_parms)]

        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)

        s = state.get('dotproduct_AE')
        self.dotproduct_AE = [IterativeDotProduct(self.dimension, iterative_shifted_mean_1 = self.iterative_shifted_mean_A, iterative_shifted_mean_2 = self.iterative_shifted_mean_E[i], state = s[i]) for i in range(self.nb_parms)]
        
        s = state.get('dotproduct_BE')
        self.dotproduct_BE = [IterativeDotProduct(self.dimension, iterative_shifted_mean_1 = self.iterative_shifted_mean_B, iterative_shifted_mean_2 = self.iterative_shifted_mean_E[i], state = s[i]) for i in range(self.nb_parms)]
        



import numpy as np

from iterative_stats.sensitivity.abstract_sensitivity import IterativeAbstractSensitivity
from iterative_stats.iterative_variance import IterativeVariance
from iterative_stats.iterative_covariance import IterativeCovariance
from iterative_stats.utils.logger import logger


class IterativeSensitivityMartinez(IterativeAbstractSensitivity):
    """
    Estimates the Sobol indices iteratively based on the Martinez formula.
    """
    def __init__(self, nb_parms: int, dim: int = 1, second_order: bool = False, state: object = None):
        super().__init__(nb_parms = nb_parms, dim = dim, second_order=second_order, state=state)
        if state is None:
            
            self.var_B = IterativeVariance(dim)
            self.var_E = [IterativeVariance(dim) for _ in range(nb_parms)]
        
            self.covData_AE = [IterativeCovariance(dim) for _ in range(nb_parms)]
            self.covData_BE = [IterativeCovariance(dim) for _ in range(nb_parms)]
        
            self.pearson_A = np.zeros((dim, nb_parms))
            self.pearson_B = np.zeros((dim, nb_parms))
        
            del self.state
       
    def _increment(self, data):

        sample_A = data[0]
        sample_B = data[1]
        sample_E = data[2:(2 + self.nb_parms)]
        self.var_B.increment(sample_B)

        for p in range(self.nb_parms):
            self.var_E[p].increment(sample_E[p])

            # update first order
            self.covData_BE[p].increment(sample_B,sample_E[p])
            var_prod = np.multiply(self.var_E[p].get_stats(), self.var_B.get_stats())
            self.pearson_B[:,p] = np.divide(self.covData_BE[p].get_stats(), np.sqrt(var_prod))

            # update last order
            self.covData_AE[p].increment(sample_A,sample_E[p])
            var_prod = np.multiply(self.var_E[p].get_stats(), self.var_A.get_stats())
            if var_prod.any() > 0 :
                self.pearson_A[:,p] = np.divide(self.covData_AE[p].get_stats(), np.sqrt(var_prod))
           

    def getFirstOrderIndices(self) :
        return self.pearson_B

    def getTotalOrderIndices(self) :
        return 1. - self.pearson_A

    def _compute_varianceI(self):
        res = np.zeros((self.dimension, self.nb_parms))
        for p in range(self.nb_parms):
            res[:,p] = np.multiply(self.pearson_B[:,p],self.var_A.get_stats())
        return res

    def _compute_VTi(self):
        return None

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['var_B'] = self.var_B.save_state()
        state['var_E'] = [self.var_E[i].save_state() for i in range(self.nb_parms)]
        state['covData_AE'] = [self.covData_AE[i].save_state() for i in range(self.nb_parms)]
        state['covData_BE'] = [self.covData_BE[i].save_state() for i in range(self.nb_parms)]
        state['pearson_A'] = self.pearson_A
        state['pearson_B'] = self.pearson_B
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.var_B = IterativeVariance(self.dimension, state=state.get('var_B', None))
        self.var_E= [IterativeVariance(self.dimension, state=s) for s in state.get('var_E', None)]
        
        self.covData_AE = [IterativeCovariance(self.dimension, state=s) for s in state.get('covData_AE', None)]
        self.covData_BE = [IterativeCovariance(self.dimension, state=s) for s in state.get('covData_BE', None)]
        
        self.pearson_A = state.get('pearson_A', None)
        self.pearson_B = state.get('pearson_B', None)



import numpy as np
import copy 

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.utils.logger import logger 

class IterativeMean(AbstractIterativeStatistics):
    """
        Iterative Mean
    """

    def increment(self, data):
        self.iteration += 1
        self.state += (data - self.state) / float(self.iteration)
        logger.debug(f'increment= {self.increment}, mean= {self.state}')


class IterativeShiftedMean(AbstractIterativeStatistics):
    """
        Iterative Mean with a shift in the data
    """
    def __init__(self, dim : int = 1, state: object = None):
        self.previous_shift = None
        super().__init__(dim, state)


    def increment(self, data, shift: np.array):
        self.iteration += 1
        if self.iteration == 1 : # Initialization
            self.state += data - shift 
        else :
            # logger.info(f'------------ data: {data}')
            self.state *= (1. - 1./self.iteration)
            self.state += (data - self.previous_shift)/self.iteration  + (self.previous_shift - shift)

        # update the shift
        self.previous_shift = copy.deepcopy(shift)

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['previous_shift'] = self.previous_shift
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.previous_shift = state.get('previous_shift', None)


import numpy as np

from iterative_stats.sensitivity.abstract_sensitivity import IterativeAbstractSensitivity
from iterative_stats.utils.logger import logger



class IterativeSensitivityJansen(IterativeAbstractSensitivity):
    """
    Estimates the Sobol indices based on the Jansen estimate
    """
    def __init__(self, nb_parms: int, dim: int = 1, second_order: bool = False, state: object = None):
        super().__init__(nb_parms = nb_parms, dim = dim, second_order=second_order, state=state)
        if state is None :
            self.AminusE = np.zeros((dim, self.nb_parms))
            self.BminusE = np.zeros((dim, self.nb_parms))
            del self.state
       
    def _increment(self, data):
        sample_A = data[0]
        sample_B = data[1]
        sample_E = data[2:(2 + self.nb_parms)]
        for p in range(self.nb_parms):
            self.AminusE[:,p] += np.multiply(sample_A- sample_E[p], sample_A - sample_E[p])
            self.BminusE[:,p] += np.multiply(sample_B- sample_E[p], sample_B - sample_E[p])
        

    def _compute_varianceI(self) :
        return self.var_A.get_stats()[:, None] - self.BminusE/(2*self.iteration - 1)

    def _compute_VTi(self) :
        coeff = 2*self.iteration - 1
        return self.AminusE/coeff

    def getIteration(self):
        return self.iteration

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['AminusE'] = self.AminusE
        state['BminusE'] = self.BminusE
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.AminusE = state.get('AminusE', None)
        self.BminusE = state.get('BminusE', None)


from abc import ABC, abstractmethod
from typing import Dict
import numpy as np
import copy

from iterative_stats.utils.logger import logger 

class AbstractExperiment(ABC):
    def __init__(self, nb_parms: int, nb_sim : int, apply_pick_freeze: bool = True, seed : int = 0, second_order: bool = False,  **kwargs) -> None:
        self.seed = seed
        self.nb_sim = nb_sim
        self.nb_parms = nb_parms
        self.apply_pick_freeze = apply_pick_freeze
        self.second_order = second_order

    def generator(self) :
        for _ in range(self.nb_sim):
            sample_A = self.draw()
            if self.apply_pick_freeze:
                sample_B = self.draw()
                sample = self.pick_freeze(sample_A[0], sample_B[0])
            else:
                sample = sample_A[0]
            yield sample

    def pick_freeze(self, sample_A: np.array, sample_B : np.array) -> np.array:
        """
            Apply the pick-freeze method and construct the set of inputs parameters
        """
        if sample_A is None :
            sample_A = self.draw()[0]
        if sample_B is None :
            sample_B = self.draw()[0]
        sample = np.vstack(([sample_A], [sample_B]))
        for k in range(self.nb_parms):
            sample_Ek = copy.deepcopy(sample_A)  
            sample_Ek[k] = sample_B[k]
            sample = np.vstack( (sample, sample_Ek))
        
        if self.second_order :
            for k in range(self.nb_parms):
                sample_Ck = copy.deepcopy(sample_B)  
                sample_Ck[k] = sample_A[k]
                sample = np.vstack( (sample, sample_Ck))
        return sample 

    @abstractmethod
    def draw(self) -> np.array:
        pass




import numpy as np

from iterative_stats.utils.logger import logger

def multi_dim_dotproduct(data_1: np.array, data_2: np.array, dim: int = 1):
    if dim == 1 :
        return np.dot(data_1, data_2)
    return np.array([np.dot(data_1[:,k], data_2[:,k]) for k in range(dim)])



import numpy as np
from typing import Dict

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.iterative_mean import IterativeShiftedMean, IterativeMean
from iterative_stats.iterative_dotproduct import IterativeDotProduct
from iterative_stats.iterative_variance import IterativeVariance
from iterative_stats.utils.logger import logger
from iterative_stats.sensitivity import SALTELLI


class IterativeAbstractSensitivity(AbstractIterativeStatistics):
    """
        Abstract class to compute the first, second and total sensitivity indices
    """

    def __init__(self, nb_parms : int, dim: int = 1, 
                        second_order: bool = False, name: str = "", state: object = None) -> None:
        """
            nb_parms (int) : number of input variables
            dim (int) : output size
            second_order (bool) : a boolean indicating if the second order must be computed or not. 
        """
        self.name : str = name
        self.nb_parms : int = nb_parms
        self.second_order : bool = second_order 

        if state is None :
            self.mean_tot : IterativeMean = IterativeMean(dim)
            self.mean_tot_iteration : int = 0
            self.var_A: AbstractIterativeStatistics = IterativeVariance(dim)

            if second_order or name == SALTELLI:
                self.iterative_shifted_mean_A = IterativeShiftedMean(dim)
                self.iterative_shifted_mean_B = IterativeShiftedMean(dim)
                self.iterative_shifted_mean_E = [IterativeShiftedMean(dim) for _ in range(nb_parms)]

            if second_order :
                self.dotproduct_AB = IterativeDotProduct(dim, iterative_shifted_mean_1 = self.iterative_shifted_mean_A, iterative_shifted_mean_2 = self.iterative_shifted_mean_B)
                self.dotproduct_EC = [[IterativeDotProduct(dim, iterative_shifted_mean_1 = self.iterative_shifted_mean_E[i]) for _ in range(nb_parms)] for i in range(nb_parms)]

        super().__init__(dim, state)


    def _increment_variance(self, data : np.array) -> None: 
        """
            data (np.array) : input data
            Function that increments the variance of sample A.
        """       
        self.var_A.increment(data[0])
        
    def getIteration(self) -> int:
        """
            Get the current iteration index.
        """
        return self.iteration

    def getFirstOrderIndices(self) -> np.array:
        """
            Compute the self.nb_parms first order sensitivity indices
        """
        if self.iteration > 1 :
            return self._compute_varianceI()/self.var_A.get_stats()[:,None]
        else :
            return None
 
    def getSecondOrderIndices(self) -> np.array:
        """
            Compute the self.nb_parms second order sensitivity indices
        """
        if self.iteration > 1 and self.second_order :   
            first_order = self._compute_varianceI()
            val = np.zeros((self.dimension, self.nb_parms, self.nb_parms))
            
            for i in range(self.nb_parms):
                for j in range(self.nb_parms):
                    val[:, i,j] = self.dotproduct_EC[i][j].get_stats() - first_order[:,i] - first_order[:,j] 
            val += self.dotproduct_AB.get_stats()[:, None, None] * (1. - 1./self.iteration)
            return val/self.var_A.get_stats()[:, None, None]
        else :
            return None

    def getTotalOrderIndices(self) -> np.array :
        """
            Compute the self.nb_parms total order sensitivity indices
        """
        if self.iteration > 1 :
            return np.divide(self._compute_VTi(),self.var_A.get_stats()[:,None])
        else :
            return None

    def increment(self, data : np.array) -> None :
        """
            Function that applies the iterative formula to increment the first and total order indices.
        """
        self.iteration += 1
        # Update the total mean
        if self.second_order or self.name == SALTELLI:
            self._update_global_mean(data)

        # Update the variance (sample A)
        self._increment_variance(data)

        # Update the specific increment data
        self._increment(data)


        if self.second_order :
            nb_required_sim = (2+ 2*self.nb_parms)
            if len(data) < nb_required_sim:
                raise Exception(f'The sample size must have {nb_required_sim} rows to compute the second order term.')
            else :
                self._increment_dotproduct(data)
        
        # Update all the shifted mean 
        if self.second_order or self.name == SALTELLI :
            self.iterative_shifted_mean_A.increment(data[0], shift = self.mean_tot.get_stats())
            self.iterative_shifted_mean_B.increment(data[1], shift = self.mean_tot.get_stats())
            for i in range(self.nb_parms):
                self.iterative_shifted_mean_E[i].increment(data[2:(2+self.nb_parms)][i], shift = self.mean_tot.get_stats())


    def _increment(self, data : np.array) -> None :
        """
            Function that applies the specific iterative formula to increment the first and total order indices.
        """
        raise Exception('Not implemented method')


    def _increment_dotproduct(self,data : np.array) -> None :
        """
            data (np.array) : a (self.nb_sim size, self.nb_parms) np.array vector
            Internal method that increments the dot products order sensitivity value.
        """

        # Update EC dot product and E iterative mean
        sample_E = data[2:2+self.nb_parms]
        sample_C = data[2+self.nb_parms:]
        for i in range(self.nb_parms) :
            for j in range(self.nb_parms):
                self.dotproduct_EC[i][j].increment(sample_E[i], sample_C[j], shift = self.mean_tot.get_stats())    
        
        # Update AB dot product
        self.dotproduct_AB.increment(data[0], data[1], shift = self.mean_tot.get_stats())
    
    def _update_global_mean(self, data : np.array) -> None :
        if self.iteration -1 == self.mean_tot_iteration : 
            # Update the global mean of the system
            for d in data :
                self.mean_tot.increment(d)
            self.mean_tot_iteration += 1 

    def _compute_varianceI(self) -> None:
        """
            Function that computes the variance V_i
        """
        raise Exception('Not implemented method')

    def _compute_VTi(self) -> None:
        """
            Function that computes the total variance V[Y] - V_-i
        """
        raise Exception('Not implemented method')

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        
        state = {}
        state['iteration'] = self.iteration
        state['mean_tot'] = self.mean_tot.save_state()
        state['mean_tot_iteration'] = self.mean_tot_iteration
        state['var_A'] = self.var_A.save_state()

        if self.second_order or self.name == SALTELLI:
            state['iterative_shifted_mean_A'] = self.iterative_shifted_mean_A.save_state()
            state['iterative_shifted_mean_B'] = self.iterative_shifted_mean_B.save_state()
            state['iterative_shifted_mean_E'] = [self.iterative_shifted_mean_E[i].save_state() for i in range(self.nb_parms)]

        if self.second_order :
            state['dotproduct_AB'] = self.dotproduct_AB.save_state()
            state['dotproduct_EC'] = [[self.dotproduct_EC[i][j].save_state() for j in range(self.nb_parms)] for i in range(self.nb_parms)]

        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        self.iteration = state.get('iteration')
        self.mean_tot = IterativeMean(self.dimension, state.get('mean_tot'))
        self.mean_tot_iteration = state.get('mean_tot_iteration')
        self.var_A = IterativeVariance(self.dimension, state.get('var_A'))

        if self.second_order or self.name == SALTELLI:
            self.iterative_shifted_mean_A = IterativeShiftedMean(self.dimension, state.get('iterative_shifted_mean_A'))
            self.iterative_shifted_mean_B = IterativeShiftedMean(self.dimension, state.get('iterative_shifted_mean_B'))
            self.iterative_shifted_mean_E = [IterativeShiftedMean(self.dimension, s) for s in state.get('iterative_shifted_mean_E')]

        if self.second_order :
            self.dotproduct_AB = IterativeDotProduct(self.dimension, 
                                                        iterative_shifted_mean_1 = self.iterative_shifted_mean_A, 
                                                        iterative_shifted_mean_2 = self.iterative_shifted_mean_B,
                                                        state= state.get('dotproduct_AB'))
            s = state.get('dotproduct_EC')
            self.dotproduct_EC = [[IterativeDotProduct(self.dimension, 
                                    iterative_shifted_mean_1 = self.iterative_shifted_mean_E[i], state=s[i][j]) for j in range(self.nb_parms)] for i in range(self.nb_parms)]

        


from typing import Dict 
import numpy as np

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.iterative_mean import IterativeMean
from iterative_stats.utils.logger import logger 

class IterativeVariance(AbstractIterativeStatistics):
    def __init__(self, dim:int = 1, state: object = None):
        if state is None :
            self.mean = IterativeMean(dim)
            self.sumOfCenteredSquares = np.zeros(dim)
        super().__init__(dim, state)

    def increment(self, data):
        self.iteration += 1
        if self.iteration > 1 : 
            self.sumOfCenteredSquares += (
                (self.iteration - 1) * (data - self.mean.get_stats()) ** 2 / self.iteration
            )

        # update mean 
        self.mean.increment(data)

        # compute variance
        if self.iteration > 1 :
            self.state = self.sumOfCenteredSquares / (self.iteration - 1)
        logger.debug(f'increment= {self.increment}, variance= {self.state}')

    def get_variance(self):
        return self.state 

    def get_mean(self):
        return self.mean.get_stats()

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['mean'] = self.mean.save_state()
        state['sumOfCenteredSquares'] = self.sumOfCenteredSquares
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.mean = IterativeMean(self.dimension, state.get('mean'))
        self.sumOfCenteredSquares = state.get('sumOfCenteredSquares')


import numpy as np
import numpy.typing as npt
from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics


class IterativeMoments(AbstractIterativeStatistics):
    """
    Iterative higher order moments following https://arxiv.org/pdf/1510.04923.pdf
    "Simpler Online Updates for Arbitrary-Order Central Moments"
    Takes:
    :param max_order: the maximum order of the moments to compute
    """
    def __init__(self, max_order: int, **kwargs):
        super().__init__(**kwargs)
        self.max_order: int = max_order

        # Cascading moments depending on the requested max_order
        self.m1: npt.NDArray[np.float_] = np.zeros(self.dimension)
        if self.max_order > 1:
            self.m2: npt.NDArray[np.float_] = np.zeros(self.dimension)
            self.theta2 = np.zeros(np.size(self.dimension))
        if self.max_order > 2:
            self.m3: npt.NDArray[np.float_] = np.zeros(self.dimension)
            self.theta3 = np.zeros(np.size(self.dimension))
        if self.max_order > 3:
            self.m4: npt.NDArray[np.float_] = np.zeros(self.dimension)
            self.theta4 = np.zeros(self.dimension)
        if self.max_order > 4:
            raise NotImplementedError("Moments of order > 4 not implemented")

    def increment(self, np_data: npt.NDArray[np.float_]):

        self.iteration += 1
        delta = self.get_delta(self.m1, np_data)
        delta_n = delta / self.iteration
        self.m1 += delta_n

        if self.max_order > 1:
            self.m2 += delta * ( delta - delta_n )
        if self.max_order > 2:
            delta_2 = delta * delta
            delta_n_2 = delta_n * delta_n
            self.m3 += -3.0 * delta_n * self.m2 + delta * ( delta_2 - delta_n_2 )
        if self.max_order > 3:
            self.m4 += -4.0 * delta_n * self.m3 - 6.0 * delta_n_2 * self.m2 \
                + delta * ( delta * delta_2 - delta_n * delta_n_2 )

    def get_delta(self, moment: npt.NDArray[np.float_], np_data):
        return np_data - moment

    def get_mean(self):
        return self.m1

    def get_variance(self):
        return self.m2 / self.iteration

    def get_skewness(self):
        return self.iteration ** 0.5 * self.m3 / self.m2 ** 1.5

    def get_kurtosis(self):
        return self.iteration * self.m4 / self.m2 ** 2

    def save_state(self):
        """
        This function saves the state of the moments
        """
        return {"m1": self.m1,
                "m2": self.m2,
                "m3": self.m3,
                "m4": self.m4,
                "increment": self.iteration}

    def load_from_state(self, state: dict):

        self.m1 = state["m1"]
        self.m2 = state["m2"]
        self.m3 = state["m3"]
        self.m4 = state["m4"]
        self.iteration = state["increment"]



import numpy as np

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.utils.logger import logger 

class IterativeExtrema(AbstractIterativeStatistics):
    def __init__(self, dim: int = 1, state: object = None):
        self.state_min = np.zeros(dim)
        super().__init__(dim, state)
        

    def increment(self, data):
        self.state_min = np.minimum(data, self.state_min)
        self.state = np.maximum(data, self.state)

    def get_min(self):
        return self.state_min

    def get_max(self):
        return self.state 

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['state_min'] = self.state_min
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.state_min = state.get('state_min', None)


from typing import Dict 
import numpy as np
import copy

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.iterative_mean import IterativeShiftedMean
from iterative_stats.utils.logger import logger 
from iterative_stats.utils.dot_prod import multi_dim_dotproduct

class IterativeDotProduct(AbstractIterativeStatistics):
    def __init__(self, dim : int = 1, iterative_shifted_mean_1 : IterativeShiftedMean = None, 
                        iterative_shifted_mean_2 : IterativeShiftedMean = None, state: object = None):
        """
            Compute iteratively the formula sum_k=0:N (A_k - shift_N)(B_k - shift_N)/(N-1)
            iterative_shifted_mean_1 (IterativeShiftedMean) : add an external iterative shifted mean. If it is not None, the mean will not be updated into this class
        """
        self.previous_shift = None

        if iterative_shifted_mean_1 is None :
            self.iterative_shifted_mean_1 = IterativeShiftedMean(dim)
            self.external_mean_1 = False
        else :
            self.iterative_shifted_mean_1 = iterative_shifted_mean_1
            self.external_mean_1 = True

        if iterative_shifted_mean_2 is None :
            self.iterative_shifted_mean_2 = IterativeShiftedMean(dim)
            self.external_mean_2 = False
        else :
            self.iterative_shifted_mean_2 = iterative_shifted_mean_2
            self.external_mean_2 = True

        super().__init__(dim, state)
        self.data_1 = None
        self.data_2 = None

    def increment(self, data_1, data_2, shift):
        self.iteration += 1
        
        if self.iteration == 1 :
            self.data_1 = data_1
            self.data_2= data_2
        elif self.iteration == 2 :
            if self.dimension == 1 :
                self.data_1 = np.array([self.data_1, data_1])
                self.data_2= np.array([self.data_2, data_2])
            else :
                self.data_1 = np.vstack((self.data_1, data_1))
                self.data_2 = np.vstack((self.data_2, data_2))
            self.state = multi_dim_dotproduct(self.data_1 - shift, self.data_2 - shift, self.dimension)
            del self.data_1 , self.data_2
        else :
            diff_shift = self.previous_shift - shift
            self.state *= (1 - 1/(self.iteration - 1))
            self.state += np.multiply(data_1 - self.previous_shift, data_2 - self.previous_shift)/(self.iteration - 1)
            val = self.iterative_shifted_mean_1.get_stats() + self.iterative_shifted_mean_2.get_stats()
            val +=  diff_shift + (data_1 + data_2 - shift - self.previous_shift)/(self.iteration - 1)
            self.state += val * diff_shift
            
            
        # Update iterative shifted mean if not external
        if not self.external_mean_1 :
            self.iterative_shifted_mean_1.increment(data_1, shift)

        if not self.external_mean_2 :
            self.iterative_shifted_mean_2.increment(data_2, shift)
        
        self.previous_shift = copy.deepcopy(shift)


    def get_mean_1(self):
        return self.iterative_shifted_mean_1.get_stats()

    def get_mean_2(self):
        return self.iterative_shifted_mean_2.get_stats()

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['previous_shift'] = self.previous_shift
        if not self.external_mean_1 :
            state['external_mean_1'] = self.iterative_shifted_mean_1.save_state()
        if not self.external_mean_2 :
            state['external_mean_2'] = self.iterative_shifted_mean_2.save_state()
        return state
        

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.previous_shift = state.get('previous_shift')
        if state.get('external_mean_1', None) is not None :
            self.external_mean_1 = False
            self.iterative_shifted_mean_1 = IterativeShiftedMean(self.dimension, state.get('external_mean_1'))
        if state.get('external_mean_2', None) is not None :
            self.external_mean_2 = False
            self.iterative_shifted_mean_2 = IterativeShiftedMean(self.dimension, state.get('external_mean_2'))



# Iterative quantile estimation by use of a Robbins-Monro algorithm
# Implementer: Alejandro Ribés

import numpy as np

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.utils.logger import logger 

class IterativeQuantile(AbstractIterativeStatistics):
    def __init__(self, dim: int = 1, state: object = None, alpha: float = 0.5, max_it: int = 0):
        if state is None :
            self.state_min = np.zeros(dim)
            self.state_max = np.zeros(dim)
            self.indicator = np.zeros(dim)
            self.C = np.zeros(dim)
        super().__init__(dim, state)

        self.max_it = max_it
        self.alpha = alpha   


    def increment(self, data):
        self.iteration += 1
        
        if self.iteration == 1:
            self.state_min = data
            self.state_max = data
        else:
            self.state_min = np.minimum(data, self.state_min)
            self.state_max = np.maximum(data, self.state_max)
            
        if self.max_it <= 1.0 :
            logger.error(f'MaxIterations= {self.max_it}, it has not been set (or set to <= 1)')

        gamma = 0.5 + 0.5 * ((self.iteration - 1.0)/(self.max_it - 1.0))
        C = abs(self.state_max - self.state_min)
        
        self.indicator = (data <= self.state).astype(np.int8)

        if self.iteration == 1:
            self.state = data
        else:
            self.state -= (C / self.iteration**gamma) * (self.indicator - self.alpha)
        

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['indicator'] = self.indicator
        state['C'] = self.C
        state['state_min'] = self.state_min
        state['state_max'] = self.state_max
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.state_min = state.get('state_min')
        self.state_max = state.get('state_max')
        self.indicator = state.get('indicator')
        self.C = state.get('C')


from typing import Dict 
import numpy as np
import copy

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.iterative_mean import IterativeMean
from iterative_stats.utils.logger import logger 

class IterativeCovariance(AbstractIterativeStatistics):
    def __init__(self, dim : int = 1, state: object = None):
        self.mean_1 = IterativeMean(dim)
        self.mean_2 = IterativeMean(dim)
        super().__init__(dim, state)

    def increment(self, data_1, data_2):
        # update mean
        self.iteration += 1
        prev_mean_1 = copy.deepcopy(self.mean_1.get_stats())
        prev_mean_2 = copy.deepcopy(self.mean_2.get_stats())
        self.mean_1.increment(data_1)
        self.mean_2.increment(data_2)
        
        # update covariance
        if self.iteration > 1:
            self.state = self.state * (self.iteration - 2)
            x = np.multiply(data_1 - prev_mean_1, data_2 - prev_mean_2)
            diff_mean = np.multiply(prev_mean_1 - self.mean_1.get_stats(), prev_mean_2 - self.mean_2.get_stats()) 
            self.state = self.state + x - diff_mean * self.iteration
            self.state = self.state / (self.iteration - 1)

    def getCovariance(self):
        return self.state

    def get_mean1(self):
        return self.mean_1.get_stats()
    
    def get_mean2(self):
        return self.mean_2.get_stats()

    def save_state(self):
        """
            An abstract method to implement. It save the current state of the objects.
        """
        state = super().save_state()
        state['mean_1'] = self.mean_1.save_state()
        state['mean_2'] = self.mean_2.save_state()
        return state

    def load_from_state(self, state: object):
        """
            It load the current state of the object.
        """
        super().load_from_state(state)
        self.mean_1 = IterativeMean(self.dimension, state=state.get('mean_1', None))
        self.mean_2 = IterativeMean(self.dimension, state=state.get('mean_2', None))


import numpy as np
import copy 

from iterative_stats.abstract_iterative_statistics import AbstractIterativeStatistics
from iterative_stats.utils.logger import logger 

class IterativeThreshold(AbstractIterativeStatistics):
    """
        Iterative Threshold
    """

    def __init__(self, dim:int = 1, min_threshold: np.array = None, max_threshold: np.array = None, state: object = None):
        super().__init__(dim, state)
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold

    def increment(self, data):
        self.iteration += 1

        if self.min_threshold is not None:
            res_min = data >= self.min_threshold
        else :
            res_min = np.ones(self.dim)
        if self.max_threshold is not None:
            res_max = data <= self.max_threshold
        else :
            res_max = np.ones(self.dim)

        self.state += res_min & res_max
        
